﻿using System.Collections.Generic;
using System.Drawing;
using System.Linq;
using HeuristicLab.Core.Views;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding.Views;
using HeuristicLab.EvolutionTracking;
using HeuristicLab.MainForm;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.Views {
  [View("FragmentGraphView")]
  [Content(typeof(IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>>), IsDefaultView = true)]
  public sealed partial class FragmentGraphView : ItemView {
    private int PreferredSpacing = 50;

    private ReingoldTilfordLayoutEngine<TileLayoutNode> layoutEngine;
    private ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode> symbolicExpressionEngine;

    private Dictionary<IGenealogyGraphNode<IFragment<ISymbolicExpressionTreeNode>>, TileLayoutNode> tileDictionary;

    private SymbolicExpressionTreeTile Root { get; set; }

    public new IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>> Content {
      get { return (IGenealogyGraph<IFragment<ISymbolicExpressionTreeNode>>)base.Content; }
      set { base.Content = value; }
    }

    public FragmentGraphView() {
      InitializeComponent();

      layoutEngine = new ReingoldTilfordLayoutEngine<TileLayoutNode>(n => n.Children);
      symbolicExpressionEngine = new ReingoldTilfordLayoutEngine<ISymbolicExpressionTreeNode>(n => n.Subtrees);
      tileDictionary = new Dictionary<IGenealogyGraphNode<IFragment<ISymbolicExpressionTreeNode>>, TileLayoutNode>();
    }

    private void MakeTiles() {
      var chart = symbolicExpressionChartControl.Chart;
      tileDictionary.Clear();
      foreach (var node in Content.Nodes) {
        var tile = new SymbolicExpressionTreeTile(chart);
        tile.LayoutEngine = symbolicExpressionEngine;
        tile.Root = node.Content.Root;
        var tileNode = new TileLayoutNode { Tile = tile };
        tileDictionary.Add(node, tileNode);
      }
      foreach (var node in Content.Nodes.Where(n => n.Children.Any())) {
        var layoutNode = tileDictionary[node];
        layoutNode.Children = new List<TileLayoutNode>(node.Children.Select(x => tileDictionary[x]));
      }
    }

    private void Draw() {
      var root = Content.Nodes.First();

      var fragmentRoot = tileDictionary[root];
      var visualNodes = layoutEngine.CalculateLayout(fragmentRoot);

      symbolicExpressionChartControl.UpdateEnabled = false;
      foreach (var visualNode in visualNodes) {
        var tile = visualNode.Content.Tile;
        tile.Position = new Point(visualNode.X, visualNode.Y);
        symbolicExpressionChartControl.Add(tile);
      }
      symbolicExpressionChartControl.UpdateEnabled = true;
      symbolicExpressionChartControl.EnforceUpdate();
    }

    protected override void DeregisterContentEvents() {
      // TODO: Deregister your event handlers here
      base.DeregisterContentEvents();
    }

    protected override void RegisterContentEvents() {
      base.RegisterContentEvents();
      // TODO: Register your event handlers here
    }

    #region Event Handlers (Content)
    // TODO: Put event handlers of the content here
    protected override void OnContentChanged() {
      base.OnContentChanged();
      if (Content != null) {
        MakeTiles();
        Draw();
      }
    }
    #endregion

    protected override void SetEnabledStateOfControls() {
      base.SetEnabledStateOfControls();
      // TODO: Enable or disable controls based on whether the content is null or the view is set readonly
    }

    #region Event Handlers (child controls)

    // TODO: Put event handlers of child controls here.

    #endregion
  }

  internal class TileLayoutNode {
    public SymbolicExpressionTreeTile Tile { get; set; }

    private List<TileLayoutNode> children;
    public IEnumerable<TileLayoutNode> Children {
      get { return children ?? Enumerable.Empty<TileLayoutNode>(); }
      set { children = value.ToList(); }
    }
  }
}